from cached_property import cached_property
from elasticsearch_dsl.query import MultiMatch, MatchPhrase, FunctionScore
from elasticsearch_dsl.function import SF
from onegov.core.collection import Pagination


class Search(Pagination):

    results_per_page = 10
    max_query_length = 100

    def __init__(self, request, query, page):
        self.request = request
        self.query = query
        self.page = page

    @cached_property
    def available_documents(self):
        search = self.request.app.es_search_by_request(self.request)
        return search.count()

    @cached_property
    def explain(self):
        return self.request.is_manager and 'explain' in self.request.params

    @property
    def q(self):
        return self.query

    def __eq__(self, other):
        return self.page == other.page and self.query == other.query

    def subset(self):
        return self.batch

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return Search(self.request, self.query, index)

    @cached_property
    def batch(self):
        if not self.query:
            return None

        search = self.request.app.es_search_by_request(
            request=self.request,
            explain=self.explain
        )

        # queries need to be cut at some point to make sure we're not
        # pushing the elasticsearch cluster to the brink
        query = self.query[:self.max_query_length]

        # make sure the title matches with a higher priority, otherwise the
        # "get lucky" functionality is not so lucky after all
        match_title = MatchPhrase(title={"query": query, "boost": 3})

        # we *could* use Match here and include '_all' fields, but that
        # yields us less exact results, probably because '_all' includes some
        # metadata fields we have no use for
        match_rest = MultiMatch(query=query, fields=[
            field for field in self.request.app.es_mappings.registered_fields
            if not field.startswith('es_')
        ], fuzziness='1')

        search = search.query(match_title | match_rest)

        # favour documents with recent changes, over documents without
        search.query = FunctionScore(query=search.query, functions=[
            SF('gauss', es_last_change={
                'offset': '7d',
                'scale': '90d',
                'decay': '0.99'
            })
        ])

        search = search[self.offset:self.offset + self.batch_size]
        return search.execute()

    def feeling_lucky(self):
        if self.batch:
            first_entry = self.batch[0].load()

            # XXX the default view to the event should be doing the redirect
            if first_entry.__tablename__ == 'events':
                return self.request.link(first_entry, 'latest')
            else:
                return self.request.link(first_entry)

    @cached_property
    def subset_count(self):
        return self.cached_subset and self.cached_subset.hits.total or 0

    def suggestions(self):
        return tuple(self.request.app.es_suggestions_by_request(
            self.request, self.query
        ))



import sedate

from cached_property import cached_property
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from dateutil.rrule import rrule, DAILY, MO, TU, WE, TH, FR, SA, SU
from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.org import _
from uuid import uuid4
from wtforms.fields import StringField, RadioField
from wtforms.fields.html5 import DateField, IntegerField
from wtforms.validators import DataRequired, NumberRange, InputRequired
from wtforms_components import TimeField


WEEKDAYS = (
    (str(MO.weekday), _("Mo")),
    (str(TU.weekday), _("Tu")),
    (str(WE.weekday), _("We")),
    (str(TH.weekday), _("Th")),
    (str(FR.weekday), _("Fr")),
    (str(SA.weekday), _("Sa")),
    (str(SU.weekday), _("Su")),
)


def choices_as_integer(choices):
    if choices is None:
        return None

    return [int(c) for c in choices]


class AllocationFormHelpers(object):

    def generate_dates(self, start, end,
                       start_time=None, end_time=None, weekdays=None):
        """ Takes the given dates and generates the date tuples.

        The ``except_for`` field will be considered if present, as will the
        ``on_holidays`` setting.

        """

        if not (start and end):
            return []

        start = start and sedate.as_datetime(start)
        end = end and sedate.as_datetime(end)

        if start == end:
            dates = (start, )
        else:
            dates = rrule(DAILY, dtstart=start, until=end, byweekday=weekdays)

        if start_time is None or end_time is None:
            return [(d, d) for d in dates if not self.is_excluded(d)]
        else:
            if end_time < start_time:
                end_offset = timedelta(days=1)
            else:
                end_offset = timedelta()

            return [
                (
                    datetime.combine(d, start_time),
                    datetime.combine(d + end_offset, end_time)
                ) for d in dates if not self.is_excluded(d)
            ]

    def combine_datetime(self, field, time_field):
        """ Takes the given date field and combines it with the given time
        field. """

        d, t = getattr(self, field).data, getattr(self, time_field).data

        if d and not t:
            return sedate.as_datetime(d)

        if not t:
            return None

        return datetime.combine(d, t)

    def is_excluded(self, date):
        return False


class AllocationRuleForm(Form):
    """ Base form form allocation rules. """

    title = StringField(
        label=_("Title"),
        description=_("General availability"),
        validators=[InputRequired()],
        fieldset=_("Rule"))

    extend = RadioField(
        label=_("Extend"),
        validators=[InputRequired()],
        fieldset=_("Rule"),
        default='daily',
        choices=(
            ('daily', _("Extend by one day at midnight")),
            ('monthly', _("Extend by one month at the end of the month")),
            ('yearly', _("Extend by one year at the end of the year"))
        ))

    @cached_property
    def rule_id(self):
        return uuid4().hex

    @cached_property
    def iteration(self):
        return 0

    @cached_property
    def last_run(self):
        None

    @property
    def rule(self):
        return {
            'id': self.rule_id,
            'title': self.title.data,
            'extend': self.extend.data,
            'options': self.options,
            'iteration': self.iteration,
            'last_run': self.last_run,
        }

    @rule.setter
    def rule(self, value):
        self.__dict__['rule_id'] = value['id']
        self.__dict__['iteration'] = value['iteration']
        self.__dict__['last_run'] = value['last_run']

        self.title.data = value['title']
        self.extend.data = value['extend']

        for k, v in value['options'].items():
            if hasattr(self, k):
                getattr(self, k).data = v

    @property
    def options(self):
        return {
            k: getattr(self, k).data for k in self._fields
            if k not in ('title', 'extend', 'csrf_token')
        }

    def apply(self, resource):
        if self.iteration == 0:
            dates = self.dates
        else:
            unit = {
                'daily': 'days',
                'monthly': 'months',
                'yearly': 'years'
            }[self.extend.data]

            start = self.end.data + timedelta(days=1)
            end = self.end.data + relativedelta(**{unit: self.iteration})

            if hasattr(self, 'start_time'):
                start_time = self.start_time.data
                end_time = self.end_time.data
            else:
                start_time = None
                end_time = None

            dates = self.generate_dates(
                start,
                end,
                start_time=start_time,
                end_time=end_time,
                weekdays=self.weekdays
            )

        data = {**(self.data or {}), 'rule': self.rule_id}

        return len(resource.scheduler.allocate(
            dates=dates,
            whole_day=self.whole_day,
            quota=self.quota,
            quota_limit=self.quota_limit,
            data=data,
            partly_available=self.partly_available,
            skip_overlapping=True
        ))


class AllocationForm(Form, AllocationFormHelpers):
    """ Baseform for all allocation forms. Allocation forms are expected
    to implement the methods above (which contain a NotImplementedException).

    Have a look at :meth:`libres.db.scheduler.Scheduler.allocate` to find out
    more about those values.

    """

    start = DateField(
        label=_("Start"),
        validators=[InputRequired()],
        fieldset=_("Date")
    )

    end = DateField(
        label=_("End"),
        validators=[InputRequired()],
        fieldset=_("Date")
    )

    except_for = MultiCheckboxField(
        label=_("Except for"),
        choices=WEEKDAYS,
        render_kw={
            'prefix_label': False,
            'class_': 'oneline-checkboxes'
        },
        fieldset=("Date")
    )

    on_holidays = RadioField(
        label=_("On holidays"),
        choices=(
            ('yes', _("Yes")),
            ('no', _("No"))
        ),
        default='yes',
        fieldset=_("Date"))

    def on_request(self):
        if not self.request.app.org.holidays:
            self.delete_field('on_holidays')

    def ensure_start_before_end(self):
        if self.start.data and self.end.data:
            if self.start.data > self.end.data:
                self.start.errors.append(_("Start date before end date"))
                return False

    @property
    def weekdays(self):
        """ The rrule weekdays derived from the except_for field. """
        exceptions = {x for x in (self.except_for.data or tuple())}
        return [int(d[0]) for d in WEEKDAYS if d[0] not in exceptions]

    @cached_property
    def exceptions(self):
        if not hasattr(self, 'request'):
            return ()

        if not self.on_holidays:
            return ()

        if self.on_holidays.data == 'yes':
            return ()

        return self.request.app.org.holidays

    def is_excluded(self, date):
        return date.date() in self.exceptions

    @property
    def dates(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def whole_day(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def partly_available(self):
        """  Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def quota(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def quota_limit(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def data(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError


class AllocationEditForm(Form, AllocationFormHelpers):
    """ Baseform for edit forms. Edit forms differ from the base allocation
    form somewhat, since they don't offer a way to generate more than
    one allocation at a time.

    The dates property is therefore expected to return a single start, end
    dates tuple.

    """

    date = DateField(
        label=_("Date"),
        validators=[InputRequired()],
        fieldset=_("Date")
    )


class Daypasses(object):

    daypasses = IntegerField(
        label=_("Daypasses"),
        validators=[
            InputRequired(),
            NumberRange(1, 999)
        ],
        fieldset=_("Daypasses")
    )

    daypasses_limit = IntegerField(
        label=_("Daypasses Limit"),
        validators=[
            InputRequired(),
            NumberRange(0, 999)
        ],
        fieldset=_("Daypasses")
    )


class DaypassAllocationForm(AllocationForm, Daypasses):

    whole_day = True
    partly_available = False
    data = None

    @property
    def quota(self):
        return self.daypasses.data

    @property
    def quota_limit(self):
        return self.daypasses_limit.data

    @property
    def dates(self):
        return self.generate_dates(
            self.start.data,
            self.end.data,
            weekdays=self.weekdays
        )


class DaypassAllocationEditForm(AllocationEditForm, Daypasses):

    whole_day = True
    partly_available = False
    data = None

    @property
    def quota(self):
        return self.daypasses.data

    @property
    def quota_limit(self):
        return self.daypasses_limit.data

    @property
    def dates(self):
        return (
            sedate.as_datetime(self.date.data),
            sedate.as_datetime(self.date.data) + timedelta(
                days=1, microseconds=-1
            )
        )

    def apply_dates(self, start, end):
        self.date.data = start.date()

    def apply_model(self, model):
        self.date.data = model.display_start().date()
        self.daypasses.data = model.quota
        self.daypasses_limit.data = model.quota_limit


class RoomAllocationForm(AllocationForm):

    as_whole_day = RadioField(
        label=_("Whole day"),
        choices=[
            ('yes', _("Yes")),
            ('no', _("No"))
        ],
        default='yes',
        fieldset=_("Time")
    )

    start_time = TimeField(
        label=_("Each starting at"),
        description=_("HH:MM"),
        validators=[InputRequired()],
        fieldset=_("Time"),
        depends_on=('as_whole_day', 'no')
    )

    end_time = TimeField(
        label=_("Each ending at"),
        description=_("HH:MM"),
        validators=[InputRequired()],
        fieldset=_("Time"),
        depends_on=('as_whole_day', 'no')
    )

    is_partly_available = RadioField(
        label=_("May be partially reserved"),
        choices=[
            ('yes', _("Yes")),
            ('no', _("No"))
        ],
        default='yes',
        fieldset=_("Options")
    )

    data = None
    quota = 1
    quota_limit = 1

    @property
    def whole_day(self):
        return self.as_whole_day.data == 'yes'

    @property
    def partly_available(self):
        return self.is_partly_available.data == 'yes'

    @property
    def dates(self):
        return self.generate_dates(
            self.start.data,
            self.end.data,
            self.start_time.data,
            self.end_time.data,
            self.weekdays
        )


class RoomAllocationEditForm(AllocationEditForm):

    as_whole_day = RadioField(
        label=_("Whole day"),
        choices=[
            ('yes', _("Yes")),
            ('no', _("No"))
        ],
        default='yes',
        fieldset=_("Time")
    )

    start_time = TimeField(
        label=_("From"),
        description=_("HH:MM"),
        validators=[DataRequired()],
        fieldset=_("Time"),
        depends_on=('as_whole_day', 'no')
    )

    end_time = TimeField(
        label=_("Until"),
        description=_("HH:MM"),
        validators=[DataRequired()],
        fieldset=_("Time"),
        depends_on=('as_whole_day', 'no')
    )

    data = None
    quota = 1
    quota_limit = 1

    @property
    def whole_day(self):
        return self.as_whole_day.data == 'yes'

    @property
    def partly_available(self):
        return self.is_partly_available.data == 'yes'

    @property
    def dates(self):
        if self.whole_day:
            return (
                sedate.as_datetime(self.date.data),
                sedate.as_datetime(self.date.data) + timedelta(
                    days=1, microseconds=-1
                )
            )
        else:
            return (
                self.combine_datetime('date', 'start_time'),
                self.combine_datetime('date', 'end_time'),
            )

    def apply_dates(self, start, end):
        self.date.data = start.date()
        self.start_time.data = start.time()
        self.end_time.data = end.time()

    def apply_model(self, model):
        self.apply_dates(model.display_start(), model.display_end())
        self.as_whole_day.data = model.whole_day and 'yes' or 'no'



import pytest

from datetime import date, datetime, time, timedelta
from dateutil.rrule import MO, WE
from onegov.core.utils import Bunch
from onegov.event import Event
from onegov.org.forms import (
    DaypassAllocationForm,
    EventForm,
    RoomAllocationEditForm,
    RoomAllocationForm,
)
from onegov.org.forms.allocation import AllocationFormHelpers
from webob.multidict import MultiDict


@pytest.mark.parametrize('form_class', [
    DaypassAllocationForm,
    RoomAllocationForm
])
def test_allocation_form_dates(form_class):
    """ Makes sure all required methods are implemented. """
    form = form_class()

    # those are customizable
    assert hasattr(form, 'dates')
    assert hasattr(form, 'whole_day')
    assert hasattr(form, 'quota')
    assert hasattr(form, 'quota_limit')
    assert hasattr(form, 'data')
    assert hasattr(form, 'partly_available')

    # those are not (yet) customizable in onegov.org
    assert not hasattr(form, 'grouped')
    assert not hasattr(form, 'raster')
    assert not hasattr(form, 'approve_manually')


def test_daypass_single_date():
    form = DaypassAllocationForm(data={
        'start': date(2015, 8, 4),
        'end': date(2015, 8, 4),
        'daypasses': 4,
        'daypasses_limit': 1
    })

    assert form.dates == [(datetime(2015, 8, 4), datetime(2015, 8, 4))]
    assert form.whole_day == True
    assert form.quota == 4
    assert form.quota_limit == 1
    assert not form.data


def test_daypass_multiple_dates():
    form = DaypassAllocationForm(data={
        'start': date(2015, 8, 4),
        'end': date(2015, 8, 8),
    })

    assert form.dates == [
        (datetime(2015, 8, 4), datetime(2015, 8, 4)),
        (datetime(2015, 8, 5), datetime(2015, 8, 5)),
        (datetime(2015, 8, 6), datetime(2015, 8, 6)),
        (datetime(2015, 8, 7), datetime(2015, 8, 7)),
        (datetime(2015, 8, 8), datetime(2015, 8, 8)),
    ]


def test_daypass_except_for():
    form = DaypassAllocationForm(data={
        'start': datetime(2015, 8, 3),
        'end': datetime(2015, 8, 6),
        'except_for': [MO.weekday, WE.weekday]
    })

    assert form.dates == [
        (datetime(2015, 8, 4), datetime(2015, 8, 4)),
        (datetime(2015, 8, 6), datetime(2015, 8, 6)),
    ]


def test_room_single_date():
    form = RoomAllocationForm(data={
        'start': date(2015, 8, 4),
        'end': date(2015, 8, 4),
        'start_time': time(12, 00),
        'end_time': time(16, 00),
        'as_whole_day': 'no'
    })

    assert form.dates == [(datetime(2015, 8, 4, 12), datetime(2015, 8, 4, 16))]
    assert form.whole_day == False
    assert form.quota == 1
    assert form.quota_limit == 1
    assert not form.data


def test_room_whole_day():
    form = RoomAllocationForm(data={
        'start': date(2015, 8, 4),
        'end': date(2015, 8, 4),
        'as_whole_day': 'yes'
    })

    assert form.dates == [(datetime(2015, 8, 4), datetime(2015, 8, 4))]
    assert form.whole_day == True
    assert form.quota == 1
    assert form.quota_limit == 1
    assert not form.data


def test_room_multiple_dates():
    form = RoomAllocationForm(data={
        'start': date(2015, 8, 4),
        'end': date(2015, 8, 8),
        'start_time': time(12, 00),
        'end_time': time(16, 00)
    })

    assert form.dates == [
        (datetime(2015, 8, 4, 12), datetime(2015, 8, 4, 16)),
        (datetime(2015, 8, 5, 12), datetime(2015, 8, 5, 16)),
        (datetime(2015, 8, 6, 12), datetime(2015, 8, 6, 16)),
        (datetime(2015, 8, 7, 12), datetime(2015, 8, 7, 16)),
        (datetime(2015, 8, 8, 12), datetime(2015, 8, 8, 16)),
    ]


def test_room_except_for():
    form = RoomAllocationForm(data={
        'start': date(2015, 8, 3),
        'end': date(2015, 8, 6),
        'start_time': time(12, 00),
        'end_time': time(16, 00),
        'except_for': [MO.weekday, WE.weekday]
    })

    assert form.dates == [
        (datetime(2015, 8, 4, 12), datetime(2015, 8, 4, 16)),
        (datetime(2015, 8, 6, 12), datetime(2015, 8, 6, 16)),
    ]


def test_generate_dates():
    helper = AllocationFormHelpers()

    assert helper.generate_dates(date(2015, 1, 1), date(2015, 1, 1)) == [
        (datetime(2015, 1, 1), datetime(2015, 1, 1))
    ]

    assert helper.generate_dates(date(2015, 1, 1), date(2015, 1, 2)) == [
        (datetime(2015, 1, 1), datetime(2015, 1, 1)),
        (datetime(2015, 1, 2), datetime(2015, 1, 2)),
    ]

    assert helper.generate_dates(
        date(2015, 1, 4), date(2015, 1, 10), weekdays=[0, 1, 2]) == [
            (datetime(2015, 1, 5), datetime(2015, 1, 5)),
            (datetime(2015, 1, 6), datetime(2015, 1, 6)),
            (datetime(2015, 1, 7), datetime(2015, 1, 7)),
    ]

    assert helper.generate_dates(date(2016, 3, 26), date(2016, 3, 28)) == [
        (datetime(2016, 3, 26), datetime(2016, 3, 26)),
        (datetime(2016, 3, 27), datetime(2016, 3, 27)),
        (datetime(2016, 3, 28), datetime(2016, 3, 28)),
    ]


def test_generate_datetimes():
    helper = AllocationFormHelpers()

    assert helper.generate_dates(
        date(2015, 1, 1), date(2015, 1, 1), time(12, 0), time(16, 0)) == [
            (datetime(2015, 1, 1, 12), datetime(2015, 1, 1, 16))
    ]

    assert helper.generate_dates(
        date(2015, 1, 1), date(2015, 1, 2), time(12, 0), time(16, 0)) == [
            (datetime(2015, 1, 1, 12), datetime(2015, 1, 1, 16)),
            (datetime(2015, 1, 2, 12), datetime(2015, 1, 2, 16))
    ]

    assert helper.generate_dates(
        date(2015, 1, 4), date(2015, 1, 10),
        time(12, 0), time(16, 0), weekdays=[0, 1, 2]) == [
            (datetime(2015, 1, 5, 12), datetime(2015, 1, 5, 16)),
            (datetime(2015, 1, 6, 12), datetime(2015, 1, 6, 16)),
            (datetime(2015, 1, 7, 12), datetime(2015, 1, 7, 16))
    ]

    assert helper.generate_dates(
        date(2015, 1, 1), date(2015, 1, 2), time(12, 0), time(0, 0)) == [
            (datetime(2015, 1, 1, 12), datetime(2015, 1, 2)),
            (datetime(2015, 1, 2, 12), datetime(2015, 1, 3))
    ]


def test_combine_datetime():
    helper = AllocationFormHelpers()

    helper.date = Bunch(data=date(2015, 1, 1))
    helper.time = Bunch(data=None)
    assert helper.combine_datetime('date', 'time') == datetime(2015, 1, 1)

    helper.time = Bunch(data=time(12, 00))
    assert helper.combine_datetime('date', 'time') == datetime(2015, 1, 1, 12)

    helper.date = Bunch(data=None)
    helper.time = Bunch(data=None)
    assert helper.combine_datetime('date', 'time') is None


def test_edit_room_allocation_form():

    form = RoomAllocationEditForm()
    form.apply_dates(datetime(2015, 1, 1, 12, 0), datetime(2015, 1, 1, 18, 0))

    assert form.date.data == date(2015, 1, 1)
    assert form.start_time.data == time(12, 00)
    assert form.end_time.data == time(18, 00)

    form.as_whole_day.data = 'no'
    assert form.dates == (
        datetime(2015, 1, 1, 12, 0),
        datetime(2015, 1, 1, 18, 0)
    )

    form.as_whole_day.data = 'yes'
    assert form.dates == (
        datetime(2015, 1, 1),
        datetime(2015, 1, 2) - timedelta(microseconds=1)
    )

    form.apply_model(Bunch(
        display_start=lambda: datetime(2015, 1, 1, 12),
        display_end=lambda: datetime(2015, 1, 1, 18),
        whole_day=False
    ))

    assert form.dates == (
        datetime(2015, 1, 1, 12, 0),
        datetime(2015, 1, 1, 18, 0)
    )


def test_event_form_update_apply():
    form = EventForm(MultiDict([
        ('description', 'Rendez-vous automnal des médecines.'),
        ('email', 'info@example.org'),
        ('end_date', ''),
        ('end_time', '18:00'),
        ('location', 'Salon du mieux-vivre à Saignelégier'),
        ('organizer', 'Société de Médecine'),
        ('start_date', '2015-06-16'),
        ('start_time', '09:30'),
        ('tags', 'Congress'),
        ('tags', 'Health'),
        ('title', 'Salon du mieux-vivre, 16e édition'),
        ('repeat', 'without')
    ]))
    assert form.validate()

    event = Event()
    form.populate_obj(event)
    form = EventForm()
    form.request = Bunch(translate=lambda txt: txt, include=lambda src: None)
    form.process(obj=event)
    assert form.data['description'] == 'Rendez-vous automnal des médecines.'
    assert form.data['email'] == 'info@example.org'
    assert form.data['end_date'] == None
    assert form.data['end_time'] == time(18, 0)
    assert form.data['location'] == 'Salon du mieux-vivre à Saignelégier'
    assert form.data['start_date'] == date(2015, 6, 16)
    assert form.data['start_time'] == time(9, 30)
    assert form.data['organizer'] == 'Société de Médecine'
    assert sorted(form.data['tags']) == ['Congress', 'Health']
    assert form.data['title'] == 'Salon du mieux-vivre, 16e édition'
    assert form.data['weekly'] == None


def test_event_form_update_after_midnight():
    form = EventForm(MultiDict([
        ('email', 'info@example.org'),
        ('end_time', '8:00'),
        ('start_date', '2015-06-16'),
        ('start_time', '09:30'),
        ('title', 'Salon du mieux-vivre, 16e édition'),
        ('organizer', 'Société de Médecine'),
        ('location', 'Salon du mieux-vivre à Saignelégier'),
        ('repeat', 'without')
    ]))
    assert form.validate()

    event = Event()
    form.populate_obj(event)
    assert event.end.day == 17


def test_event_form_validate_weekly():
    form = EventForm(MultiDict([
        ('email', 'info@example.org'),
        ('end_date', '2015-06-23'),
        ('end_time', '18:00'),
        ('start_date', '2015-06-16'),
        ('start_time', '09:30'),
        ('title', 'Salon du mieux-vivre, 16e édition'),
        ('location', 'Salon du mieux-vivre à Saignelégier'),
        ('organizer', 'Société de Médecine'),
        ('weekly', 'MO'),
        ('repeat', 'weekly')
    ]))
    form.request = Bunch(translate=lambda txt: txt, include=lambda src: None)

    assert not form.validate()
    assert form.errors == {
        'weekly': ['The weekday of the start date must be selected.']
    }

    form = EventForm(MultiDict([
        ('email', 'info@example.org'),
        ('end_date', ''),
        ('end_time', '18:00'),
        ('start_date', '2015-06-16'),
        ('start_time', '09:30'),
        ('title', 'Salon du mieux-vivre, 16e édition'),
        ('location', 'Salon du mieux-vivre à Saignelégier'),
        ('organizer', 'Société de Médecine'),
        ('weekly', 'TU'),
        ('repeat', 'weekly')
    ]))
    form.request = Bunch(translate=lambda txt: txt, include=lambda src: None)

    assert not form.validate()
    assert form.errors == {
        'end_date': ['Please set and end date if the event is recurring.']
    }

    form = EventForm(MultiDict([
        ('email', 'info@example.org'),
        ('end_date', '2015-06-23'),
        ('end_time', '18:00'),
        ('start_date', '2015-06-16'),
        ('start_time', '09:30'),
        ('title', 'Salon du mieux-vivre, 16e édition'),
        ('location', 'Salon du mieux-vivre à Saignelégier'),
        ('organizer', 'Société de Médecine'),
        ('repeat', 'weekly')
    ]))
    form.request = Bunch(translate=lambda txt: txt, include=lambda src: None)

    assert not form.validate()
    assert form.errors == {
        'weekly': ['Please select a weekday if the event is recurring.']
    }


def test_event_form_create_rrule():

    def occurrences(form):
        event = Event()
        form.populate_obj(event)
        return [occurrence.date() for occurrence in event.occurrence_dates()]

    form = EventForm(data={
        'start_time': time(9, 30),
        'end_time': time(18, 0),
        'start_date': date(2015, 6, 1),
        'end_date': date(2015, 6, 7),
        'weekly': ['MO'],
        'tags': [],
        'repeat': 'weekly'
    })
    assert occurrences(form) == [date(2015, 6, 1)]

    form.end_date.data = date(2015, 6, 8)
    assert occurrences(form) == [date(2015, 6, 1), date(2015, 6, 8)]

    form.end_date.data = date(2015, 6, 15)
    assert occurrences(form) == [date(2015, 6, day) for day in (1, 8, 15)]

    form.end_date.data = date(2015, 6, 10)
    form.weekly.data = ['MO', 'WE']
    assert occurrences(form) == [date(2015, 6, day) for day in (1, 3, 8, 10)]

    form.start_date.data = date(2015, 6, 3)
    form.end_date.data = date(2015, 6, 10)
    form.weekly.data = ['MO', 'WE', 'FR']
    assert occurrences(form) == [date(2015, 6, day) for day in (3, 5, 8, 10)]



import morepath

from datetime import timedelta
from libres.db.models import ReservedSlot
from libres.modules.errors import LibresError
from onegov.core.security import Public, Private, Secret
from onegov.core.utils import is_uuid
from onegov.form import merge_forms
from onegov.org import OrgApp, utils, _
from onegov.org.forms import AllocationRuleForm
from onegov.org.forms import DaypassAllocationEditForm
from onegov.org.forms import DaypassAllocationForm
from onegov.org.forms import RoomAllocationEditForm
from onegov.org.forms import RoomAllocationForm
from onegov.org.layout import AllocationEditFormLayout
from onegov.org.layout import AllocationRulesLayout
from onegov.org.layout import ResourceLayout
from onegov.org.new_elements import Link, Confirm, Intercooler
from onegov.reservation import Allocation
from onegov.reservation import Reservation
from onegov.reservation import Resource
from onegov.reservation import ResourceCollection
from purl import URL
from sedate import utcnow
from sqlalchemy import not_, func
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.orm import defer, defaultload
from webob import exc


@OrgApp.json(model=Resource, name='slots', permission=Public)
def view_allocations_json(self, request):
    """ Returns the allocations in a fullcalendar compatible events feed.

    See `<http://fullcalendar.io/docs/event_data/events_json_feed/>`_ for
    more information.

    """

    start, end = utils.parse_fullcalendar_request(request, self.timezone)

    if not (start and end):
        return tuple()

    # get all allocations (including mirrors), for the availability calculation
    query = self.scheduler.allocations_in_range(start, end, masters_only=False)
    query = query.order_by(Allocation._start)
    query = query.options(defer(Allocation.data))
    query = query.options(defer(Allocation.group))
    query = query.options(
        defaultload('reserved_slots')
        .defer('reservation_token')
        .defer('allocation_id')
        .defer('end'))

    allocations = query.all()

    # but only return the master allocations
    return tuple(
        e.as_dict() for e in utils.AllocationEventInfo.from_allocations(
            request, self.scheduler, allocations
        )
    )


@OrgApp.view(model=Resource, name='process-rules', permission=Secret)
def process_rules(self, request):
    """ Manually runs the rules processing cronjobs for testing.

    Not really dangerous, though it should be replaced with something
    proper instead, cronjobs currently do not run in most tests and that
    should be remedied.

    """

    if request.current_username == 'admin@example.org':
        handle_rules_cronjob(self, request)


@OrgApp.html(model=Resource, name='rules', permission=Private,
             template='allocation_rules.pt')
def view_allocation_rules(self, request):
    layout = AllocationRulesLayout(self, request)

    def link_for_rule(rule, name):
        url = URL(request.link(self, name))
        url = url.query_param('csrf-token', layout.csrf_token)
        url = url.query_param('rule', rule['id'])

        return url.as_string()

    def actions_for_rule(rule):
        yield Link(
            text=_("Stop"),
            url=link_for_rule(rule, 'stop-rule'),
            traits=(
                Confirm(
                    _(
                        'Do you really want to stop "${title}"?',
                        mapping={'title': rule['title']}
                    ),
                    _(
                        "The rule will be removed without affecting "
                        "existing allocations."
                    ),
                    _("Stop rule")
                ),
                Intercooler(
                    request_method="POST",
                    redirect_after=request.link(self, 'rules')
                )
            )
        )

        yield Link(
            text=_("Delete"),
            url=link_for_rule(rule, 'delete-rule'),
            traits=(
                Confirm(
                    _(
                        'Do you really want to delete "${title}"?',
                        mapping={'title': rule['title']}
                    ),
                    _(
                        "All allocations created by the rule will be removed, "
                        "if they haven't been reserved yet."
                    ),
                    _("Delete rule")
                ),
                Intercooler(
                    request_method="POST",
                    redirect_after=request.link(self, 'rules')
                )
            )
        )

    def rules_with_actions():
        form_class = get_allocation_rule_form_class(self, request)

        for rule in self.content.get('rules', ()):
            form = request.get_form(form_class, csrf_support=False, model=self)
            form.rule = rule

            yield {
                'title': rule['title'],
                'actions': tuple(actions_for_rule(rule)),
                'form': form
            }

    return {
        'layout': layout,
        'title': _("Rules"),
        'rules': tuple(rules_with_actions())
    }


def get_new_allocation_form_class(resource, request):
    """ Returns the form class for new allocations (different resources have
    different allocation forms).

    """

    if resource.type == 'daypass':
        return DaypassAllocationForm

    if resource.type == 'room':
        return RoomAllocationForm

    raise NotImplementedError


def get_edit_allocation_form_class(allocation, request):
    """ Returns the form class for existing allocations (different resources
    have different allocation forms).

    """

    resource = ResourceCollection(
        request.app.libres_context).by_id(allocation.resource)

    if resource.type == 'daypass':
        return DaypassAllocationEditForm

    if resource.type == 'room':
        return RoomAllocationEditForm

    raise NotImplementedError


def get_allocation_rule_form_class(resource, request):
    """ Returns the form class for allocation rules. """

    form = get_new_allocation_form_class(resource, request)

    return merge_forms(AllocationRuleForm, form)


@OrgApp.form(model=Resource, template='form.pt', name='new-allocation',
             permission=Private, form=get_new_allocation_form_class)
def handle_new_allocation(self, request, form):
    """ Handles new allocations for differing form classes. """

    if form.submitted(request):
        try:
            allocations = self.scheduler.allocate(
                dates=form.dates,
                whole_day=form.whole_day,
                quota=form.quota,
                quota_limit=form.quota_limit,
                data=form.data,
                partly_available=form.partly_available
            )
        except LibresError as e:
            utils.show_libres_error(e, request)
        else:
            if not allocations:
                request.alert(_("No allocations to add"))
            else:
                request.success(
                    _("Successfully added ${n} allocations", mapping={
                        'n': len(allocations)
                    }))

                self.highlight_allocations(allocations)

                return morepath.redirect(request.link(self))

    elif not request.POST:
        start, end = utils.parse_fullcalendar_request(request, self.timezone)
        whole_day = request.params.get('whole_day') == 'yes'

        if start and end:
            if whole_day:
                form.start.data = start
                form.end.data = end

                if hasattr(form, 'as_whole_day'):
                    form.as_whole_day.data = 'yes'
            else:
                form.start.data = start
                form.end.data = end

                if hasattr(form, 'as_whole_day'):
                    form.as_whole_day.data = 'no'

                if hasattr(form, 'start_time'):
                    form.start_time.data = start
                    form.end_time.data = end

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("New allocation"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("New allocation"),
        'form': form
    }


@OrgApp.form(model=Allocation, template='form.pt', name='edit',
             permission=Private, form=get_edit_allocation_form_class)
def handle_edit_allocation(self, request, form):
    """ Handles edit allocation for differing form classes. """

    resources = ResourceCollection(request.app.libres_context)
    resource = resources.by_id(self.resource)

    # this is a bit of a hack to keep the current view when a user drags an
    # allocation around, which opens this form and later leads back to the
    # calendar - if the user does this on the day view we want to return to
    # the same day view after the process
    # therefore we set the view on the resource (where this is okay) and on
    # the form action (where it's a bit of a hack), to ensure that the view
    # parameter is around for the whole time
    if 'view' in request.params:
        resource.view = view = request.params['view']
        form.action = URL(form.action).query_param('view', view).as_string()

    if form.submitted(request):
        new_start, new_end = form.dates

        try:
            resource.scheduler.move_allocation(
                master_id=self.id,
                new_start=new_start,
                new_end=new_end,
                new_quota=form.quota,
                quota_limit=form.quota_limit
            )
        except LibresError as e:
            utils.show_libres_error(e, request)
        else:
            # when we edit an allocation, we disassociate it from any rules
            if self.data and 'rule' in self.data:
                self.data = {k: v for k, v in self.data.items() if k != 'rule'}

            request.success(_("Your changes were saved"))
            resource.highlight_allocations([self])

            return morepath.redirect(request.link(resource))
    elif not request.POST:
        form.apply_model(self)

        start, end = utils.parse_fullcalendar_request(request, self.timezone)
        if start and end:
            form.apply_dates(start, end)

    return {
        'layout': AllocationEditFormLayout(self, request),
        'title': _("Edit allocation"),
        'form': form
    }


@OrgApp.view(model=Allocation, request_method='DELETE', permission=Private)
def handle_delete_allocation(self, request):
    """ Deletes the given resource (throwing an error if there are existing
    reservations associated with it).

    """
    request.assert_valid_csrf_token()

    resource = request.app.libres_resources.by_allocation(self)
    resource.scheduler.remove_allocation(id=self.id)

    @request.after
    def trigger_calendar_update(response):
        response.headers.add('X-IC-Trigger', 'rc-allocations-changed')


@OrgApp.form(model=Resource, template='form.pt', name='new-rule',
             permission=Private, form=get_allocation_rule_form_class)
def handle_allocation_rule(self, request, form):
    layout = AllocationRulesLayout(self, request)

    if form.submitted(request):
        changes = form.apply(self)

        rules = self.content.get('rules', [])
        rules.append(form.rule)
        self.content['rules'] = rules

        request.success(_(
            "New rule active, ${n} allocations created", mapping={'n': changes}
        ))

        return request.redirect(request.link(self, name='rules'))

    return {
        'layout': layout,
        'title': _("New Rule"),
        'form': form,
        'helptext': _(
            "Rules ensure that the allocations between start/end exist and "
            "that they are extended beyond those dates at the given "
            "intervals. "
        )
    }


def rule_id_from_request(request):
    """ Returns the rule_id from the request params, ensuring that
    an actual uuid is returned.

    """
    rule_id = request.params.get('rule')

    if not is_uuid(rule_id):
        raise exc.HTTPBadRequest()

    return rule_id


def handle_rules_cronjob(resource, request):
    """ Handles all cronjob duties of the rules stored on the given
    resource.

    """
    if not resource.content.get('rules'):
        return

    targets = []

    now = utcnow()
    tomorrow = (now + timedelta(days=1)).date()

    def should_process(rule):
        # do not reprocess rules if they were processed less than 12 hours
        # prior - this prevents flaky cronjobs from accidentally processing
        # rules too often
        if rule['last_run']\
                and rule['last_run'] > utcnow() - timedelta(hours=12):
            return False

        # we assume to be called once a day, so if we are called, a daily
        # rule has to be processed
        if rule['extend'] == 'daily':
            return True

        if rule['extend'] == 'monthly' and tomorrow.day == 1:
            return True

        if rule['extend'] == 'yearly'\
                and tomorrow.month == 1\
                and tomorrow.day == 1:

            return True

        return False

    def prepare_rule(rule):
        if should_process(rule):
            rule['iteration'] += 1
            rule['last_run'] = now

            targets.append(rule)

        return rule

    resource.content['rules'] = [
        prepare_rule(r) for r
        in resource.content.get('rules', ())]

    form_class = get_allocation_rule_form_class(resource, request)

    for rule in targets:
        form = request.get_form(form_class, csrf_support=False, model=resource)
        form.rule = rule
        form.apply(resource)


def delete_rule(resource, rule_id):
    """ Removes the given rule from the resource. """

    resource.content['rules'] = [
        rule for rule in resource.content.get('rules', ())
        if rule['id'] != rule_id
    ]


@OrgApp.view(model=Resource, request_method='POST', permission=Private,
             name='stop-rule')
def handle_stop_rule(self, request):
    request.assert_valid_csrf_token()

    rule_id = rule_id_from_request(request)
    delete_rule(self, rule_id)

    request.success(_("The rule was stopped"))


@OrgApp.view(model=Resource, request_method='POST', permission=Private,
             name='delete-rule')
def handle_delete_rule(self, request):
    request.assert_valid_csrf_token()

    rule_id = rule_id_from_request(request)

    # all the slots
    slots = self.scheduler.managed_reserved_slots()
    slots = slots.with_entities(ReservedSlot.allocation_id)

    # all the reservations
    reservations = self.scheduler.managed_reservations()
    reservations = reservations.with_entities(Reservation.target)

    # include the allocations created by the given rule...
    candidates = self.scheduler.managed_allocations()
    candidates = candidates.filter(
        func.json_extract_path_text(
            func.cast(Allocation.data, JSON), 'rule'
        ) == rule_id
    )

    # .. without the ones with slots
    candidates = candidates.filter(
        not_(Allocation.id.in_(slots.subquery())))

    # .. without the ones with reservations
    candidates = candidates.filter(
        not_(Allocation.group.in_(reservations.subquery())))

    # delete the allocations
    count = candidates.delete('fetch')

    delete_rule(self, rule_id)

    request.success(
        _("The rule was deleted, along with ${n} allocations", mapping={
            'n': count
        })
    )



from onegov.form.extensions import FormExtension
from onegov.gis import CoordinatesField
from onegov.org import _
from wtforms.fields.html5 import EmailField
from wtforms.validators import DataRequired


class CoordinatesFormExtension(FormExtension, name='coordinates'):

    def create(self):
        class CoordinatesForm(self.form_class):
            coordinates = CoordinatesField(
                label=_("Coordinates"),
                description=_(
                    "The marker can be moved by dragging it with the mouse"
                ),
                fieldset=_("Map"),
                render_kw={'data-map-type': 'marker'}
            )

        return CoordinatesForm


class SubmitterFormExtension(FormExtension, name='submitter'):

    def create(self):
        class SubmitterForm(self.form_class):
            submitter = EmailField(
                label=_("E-Mail"),
                fieldset=_("Submitter"),
                validators=[DataRequired()]
            )

        return SubmitterForm



from onegov.org.models.atoz import AtoZ
from onegov.org.models.clipboard import Clipboard
from onegov.org.models.directory import DirectorySubmissionAction
from onegov.org.models.directory import ExtendedDirectory
from onegov.org.models.directory import ExtendedDirectoryEntry
from onegov.org.models.editor import Editor
from onegov.org.models.export import Export, ExportCollection
from onegov.org.models.extensions import ContactExtension
from onegov.org.models.extensions import ContentExtension
from onegov.org.models.extensions import CoordinatesExtension
from onegov.org.models.extensions import HiddenFromPublicExtension
from onegov.org.models.extensions import PersonLinkExtension
from onegov.org.models.extensions import VisibleOnHomepageExtension
from onegov.org.models.file import GeneralFile
from onegov.org.models.file import GeneralFileCollection
from onegov.org.models.file import ImageFile
from onegov.org.models.file import ImageFileCollection
from onegov.org.models.file import ImageSet
from onegov.org.models.file import ImageSetCollection
from onegov.org.models.form import BuiltinFormDefinition, CustomFormDefinition
from onegov.org.models.legacy_file import LegacyFile
from onegov.org.models.legacy_file import LegacyFileCollection
from onegov.org.models.legacy_file import LegacyImage
from onegov.org.models.legacy_file import LegacyImageCollection
from onegov.org.models.message import DirectoryMessage
from onegov.org.models.message import EventMessage
from onegov.org.models.message import PaymentMessage
from onegov.org.models.message import ReservationMessage
from onegov.org.models.message import SubmissionMessage
from onegov.org.models.message import TicketMessage
from onegov.org.models.message import TicketNote
from onegov.org.models.organisation import Organisation
from onegov.org.models.page import AtoZPages, News, Topic
from onegov.org.models.page_move import PageMove
from onegov.org.models.person_move import FormPersonMove
from onegov.org.models.person_move import PagePersonMove
from onegov.org.models.person_move import PersonMove
from onegov.org.models.person_move import ResourcePersonMove
from onegov.org.models.publication import PublicationCollection
from onegov.org.models.recipient import ResourceRecipient
from onegov.org.models.recipient import ResourceRecipientCollection
from onegov.org.models.resource import DaypassResource
from onegov.org.models.search import Search
from onegov.org.models.sitecollection import SiteCollection
from onegov.org.models.swiss_holidays import SwissHolidays
from onegov.org.models.traitinfo import TraitInfo

__all__ = [
    'AtoZ',
    'AtoZPages',
    'BuiltinFormDefinition',
    'Clipboard',
    'ContactExtension',
    'ContentExtension',
    'CoordinatesExtension',
    'CustomFormDefinition',
    'DaypassResource',
    'DirectoryMessage',
    'DirectorySubmissionAction',
    'Editor',
    'EventMessage',
    'Export',
    'ExportCollection',
    'ExtendedDirectory',
    'ExtendedDirectoryEntry',
    'FormPersonMove',
    'GeneralFile',
    'GeneralFileCollection',
    'HiddenFromPublicExtension',
    'ImageFile',
    'ImageFileCollection',
    'ImageSet',
    'ImageSetCollection',
    'LegacyFile',
    'LegacyFileCollection',
    'LegacyImage',
    'LegacyImageCollection',
    'News',
    'Organisation',
    'PageMove',
    'PagePersonMove',
    'PaymentMessage',
    'PersonLinkExtension',
    'PersonMove',
    'PublicationCollection',
    'ReservationMessage',
    'ResourcePersonMove',
    'ResourceRecipient',
    'ResourceRecipientCollection',
    'Search',
    'SiteCollection',
    'SubmissionMessage',
    'SwissHolidays',
    'TicketMessage',
    'TicketNote',
    'Topic',
    'TraitInfo',
    'VisibleOnHomepageExtension',
]



""" The onegov org collection of files uploaded to the site. """

import datetime
import isodate
import morepath
import sedate

from babel.core import Locale
from babel.dates import parse_pattern
from dateutil.parser import parse
from functools import lru_cache
from itertools import groupby
from onegov.core.filestorage import view_filestorage_file
from onegov.core.security import Private, Public
from onegov.core.templates import render_macro
from onegov.file import File, FileCollection
from onegov.file.utils import extension_for_content_type
from onegov.file.errors import AlreadySignedError, InvalidTokenError
from onegov.org import _, OrgApp
from onegov.org.new_elements import Link
from onegov.org.layout import DefaultLayout
from onegov.org.models import (
    GeneralFile,
    GeneralFileCollection,
    ImageFile,
    ImageFileCollection,
    ImageSetCollection,
    LegacyFile,
    LegacyImage,
)
from onegov.org import utils
from onegov.user import UserCollection
from sedate import to_timezone, utcnow, standardize_date
from time import time
from webob import exc
from uuid import uuid4


def get_thumbnail_size(image):
    if 'thumbnail_small' in image.reference:
        return image.reference.thumbnail_small['size']
    else:
        return ('256', '256')


class Img(object):
    """ Represents an img element. """

    __slots__ = ['src', 'alt', 'title', 'url', 'extra', 'width', 'height']

    def __init__(self, src, alt=None, title=None, url=None, extra=None,
                 width=None, height=None):
        #: The src of the image
        self.src = src

        #: The text for people that can't or won't look at the picture
        self.alt = alt

        #: The title of the image
        self.title = title

        #: The target of this image
        self.url = url

        #: The width of the image in pixel
        self.width = width

        #: The height of the image in pixel
        self.height = height

        #: Extra parameters
        self.extra = extra

    @classmethod
    def from_image(cls, layout, image):
        request = layout.request

        return cls(
            src=request.class_link(File, {'id': image.id}, 'thumbnail'),
            url=request.class_link(File, {'id': image.id}),
            alt=(image.note or '').strip(),
            width=get_thumbnail_size(image)[0],
            height=get_thumbnail_size(image)[1],
            extra=layout.csrf_protected_url(request.link(image, 'note'))
        )


@OrgApp.html(model=GeneralFileCollection, template='files.pt',
             permission=Private)
def view_get_file_collection(self, request):
    request.include('common')
    request.include('upload')
    request.include('prompt')

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Files"), '#')
    ]

    files = tuple(self.files)

    # XXX build somewhat manually for more speed
    locale = Locale.parse(request.locale)
    pattern = parse_pattern(layout.datetime_format)

    @lru_cache(maxsize=len(files) // 4)
    def format_date(date):
        date = to_timezone(date, layout.timezone)
        return pattern.apply(date, locale)

    grouped = tuple(
        (group, tuple(files))
        for group, files in groupby(files, key=lambda f: self.group(f))
    )

    return {
        'layout': layout,
        'title': _('Files'),
        'grouped': grouped,
        'count': len(files),
        'format_date': format_date,
        'model': self,
        'extension': lambda f: extension_for_content_type(
            f.content_type,
            f.name
        ),
        'actions_url': lambda file_id: request.class_link(
            GeneralFile, name="details", variables={'id': file_id}
        ),
        'upload_url': layout.csrf_protected_url(
            request.link(self, name='upload')
        )
    }


@OrgApp.html(model=GeneralFile, permission=Private, name='details')
def view_file_details(self, request):
    layout = DefaultLayout(self, request)
    extension = extension_for_content_type(
        self.reference.content_type,
        self.reference.filename
    )
    color = utils.get_extension_color(extension)

    # IE 11 caches all ajax requests otherwise
    @request.after
    def must_revalidate(response):
        response.headers.add('cache-control', 'must-revalidate')
        response.headers.add('cache-control', 'no-cache')
        response.headers.add('cache-control', 'no-store')
        response.headers['pragma'] = 'no-cache'
        response.headers['expires'] = '0'

    return render_macro(
        layout.macros['file-details'],
        request,
        {
            'id': uuid4().hex,
            'file': self,
            'layout': layout,
            'extension': extension,
            'color': color
        }
    )


@OrgApp.view(model=GeneralFile, permission=Private, name='publish',
             request_method='POST')
def handle_publish(self, request):
    request.assert_valid_csrf_token()
    self.published = True
    self.publish_date = None


@OrgApp.view(model=GeneralFile, permission=Private, name='unpublish',
             request_method='POST')
def handle_unpublish(self, request):
    request.assert_valid_csrf_token()
    self.published = False


@OrgApp.view(model=GeneralFile, permission=Private, name='update-publish-date',
             request_method='POST')
def handle_update_publish_date(self, request):
    request.assert_valid_csrf_token()
    layout = DefaultLayout(self, request)

    if request.params.get('clear', None):
        self.publish_date = None
        return

    try:
        date = parse(request.params['date'], dayfirst=True)
    except (ValueError, KeyError):
        date = self.publish_date and self.publish_date.date()
        date = date or layout.today()

    try:
        hour = next(map(int, request.params.get('hour').split(':')))
    except ValueError:
        hour = self.publish_date and self.publish_date.hour
        hour = hour or 0

    publish_date = datetime.datetime.combine(date, datetime.time(hour, 0))
    publish_date = standardize_date(publish_date, layout.timezone)

    self.publish_date = publish_date


@OrgApp.html(model=ImageFileCollection, template='images.pt',
             permission=Private)
def view_get_image_collection(self, request):
    request.include('common')
    request.include('upload')
    request.include('editalttext')

    layout = DefaultLayout(self, request)

    images = view_get_image_collection_json(
        self, request, produce_image=lambda image: Img.from_image(
            layout, image
        )
    )

    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Images"), request.link(self))
    ]

    layout.editbar_links = [
        Link(
            text=_("Manage Photo Albums"),
            url=request.class_link(ImageSetCollection),
            attrs={'class': 'new-photo-album'}
        )
    ]

    return {
        'layout': layout,
        'title': _('Images'),
        'images': images,
        'upload_url': layout.csrf_protected_url(
            request.link(self, name='upload')
        )
    }


@OrgApp.json(model=GeneralFileCollection, permission=Private, name='json')
def view_get_file_collection_json(self, request):
    return [
        {
            'link': request.class_link(File, {'id': id}),
            'title': name
        }
        for id, name in self.query().with_entities(File.id, File.name).all()
    ]


@OrgApp.json(model=ImageFileCollection, permission=Private, name='json')
def view_get_image_collection_json(self, request, produce_image=None):

    if not produce_image:
        def produce_image(image):
            return {
                'thumb': request.class_link(
                    File, {'id': image.id}, 'thumbnail'),
                'image': request.class_link(
                    File, {'id': image.id})
            }

    return [
        {
            'group': request.translate(group),
            'images': tuple(produce_image(image) for group, image in items)
        } for group, items in self.grouped_by_date(id_only=False)
    ]


def handle_file_upload(self, request):
    """ Stores the file given with the request and returns the new file object.

    """

    file = self.add(
        filename=request.params['file'].filename,
        content=request.params['file'].file
    )

    if self.supported_content_types is not 'all':
        if file.reference.content_type not in self.supported_content_types:
            raise exc.HTTPUnsupportedMediaType()

    return file


@OrgApp.view(model=FileCollection, name='upload',
             request_method='POST', permission=Private)
def view_upload_file(self, request, return_file=False):
    request.assert_valid_csrf_token()

    try:
        uploaded_file = handle_file_upload(self, request)
    except FileExistsError as e:
        # mark existing files as modified to put them in front of the queue
        uploaded_file = e.args[0]

    if return_file:
        return uploaded_file

    return morepath.redirect(request.link(self))


@OrgApp.html(model=GeneralFileCollection, name='upload',
             request_method='POST', permission=Private)
def view_upload_general_file(self, request):
    uploaded_file = view_upload_file(self, request, return_file=True)
    layout = DefaultLayout(self, request)

    return render_macro(layout.macros['file-info'], request, {
        'file': uploaded_file,
        'format_date': lambda dt: layout.format_date(dt, 'datetime'),
        'actions_url': lambda file_id: request.class_link(
            GeneralFile, name="details", variables={'id': file_id}
        ),
        'extension': lambda file: extension_for_content_type(
            file.reference.content_type,
            file.name
        )
    })


@OrgApp.html(model=ImageFileCollection, name='upload',
             request_method='POST', permission=Private)
def view_upload_image_file(self, request):
    uploaded_file = view_upload_file(self, request, return_file=True)
    layout = DefaultLayout(self, request)

    return render_macro(layout.macros['uploaded_image'], request, {
        'image': Img.from_image(layout, uploaded_file),
        'index': int(time() * 1000),
        'layout': layout
    })


@OrgApp.json(model=FileCollection, name='upload.json',
             request_method='POST', permission=Private)
def view_upload_file_by_json(self, request):
    request.assert_valid_csrf_token()

    try:
        f = handle_file_upload(self, request)
    except FileExistsError as e:
        # mark existing files as modified to put them in front of the queue
        e.args[0].modified = utcnow()

        return {
            'filelink': request.link(e.args[0]),
            'filename': e.args[0].name
        }
    except exc.HTTPUnsupportedMediaType:
        return {
            'error': True,
            'message': request.translate(_("This file type is not supported"))
        }
    except exc.HTTPRequestHeaderFieldsTooLarge:
        return {
            'error': True,
            'message': request.translate(_("The file name is too long"))
        }

    return {
        'filelink': request.link(f),
        'filename': f.name,
    }


@OrgApp.html(model=GeneralFileCollection, name='digest', permission=Public)
def view_file_digest(self, request):
    name = request.params.get('name')
    digest = request.params.get('digest')

    if not name:
        raise exc.HTTPBadRequest("missing filename")

    if not digest:
        raise exc.HTTPBadRequest("missing digest")

    metadata = self.locate_signature_metadata(digest)
    layout = DefaultLayout(self, request)

    return render_macro(layout.macros['digest_result'], request, {
        'layout': layout,
        'name': name,
        'is_known': metadata and True or False,
        'date': metadata and sedate.replace_timezone(
            isodate.parse_datetime(
                metadata['timestamp']
            ), 'UTC'
        ),
    })


@OrgApp.html(model=File, name='sign', request_method='POST',
             permission=Private)
def handle_sign(self, request):
    request.assert_valid_csrf_token()
    token = request.params.get('token')

    user = UserCollection(request.session).by_username(
        request.current_username)

    def may_sign():
        if not token:
            request.alert(_("Please submit your yubikey"))
            return False

        if not user.second_factor:
            request.alert(_("Your account is not linked to a Yubikey"))
            return False

        if not token.startswith(user.second_factor['data']):
            request.alert(_("The used Yubikey is not linked to your account"))
            return False

        return True

    try:
        if may_sign():
            request.app.sign_file(
                file=self,
                signee=request.current_username,
                token=token)

    except AlreadySignedError:
        request.alert(_("This file already has a digital signature"))
    except InvalidTokenError:
        request.alert(_("Your Yubikey could not be validated"))

    layout = DefaultLayout(self, request)

    return render_macro(layout.macros['sign_result'], request, {
        'layout': layout,
        'file': self,
    })


@OrgApp.view(model=LegacyFile, permission=Public)
@OrgApp.view(model=LegacyImage, permission=Public)
def view_old_files_redirect(self, request):
    """ Redirects to the migrated depot file if possible. As a result, old
    image urls are preserved and will continue to function.

    """

    alternate_path = self.path + '.r'

    if request.app.filestorage.isfile(alternate_path):
        with request.app.filestorage.open(alternate_path, 'r') as f:
            id = f.read()

        if isinstance(self, LegacyFile):
            file_class = GeneralFile
        else:
            file_class = ImageFile

        return exc.HTTPMovedPermanently(
            location=request.class_link(file_class, {'id': id}))

    return view_filestorage_file(self, request)



from collections import defaultdict
from datetime import date, datetime
from dateutil.easter import easter
from dateutil.relativedelta import MO, TH, FR
from dateutil.relativedelta import relativedelta as rd
from onegov.org import _


CANTONS = {
    'AG': _('Aargau'),
    'AR': _('Appenzell Ausserrhoden'),
    'AI': _('Appenzell Innerrhoden'),
    'BL': _('Basel-Landschaft'),
    'BS': _('Basel-Stadt'),
    'BE': _('Berne'),
    'FR': _('Fribourg'),
    'GE': _('Geneva'),
    'GL': _('Glarus'),
    'GR': _('Grisons'),
    'JU': _('Jura'),
    'LU': _('Lucerne'),
    'NE': _('Neuchâtel'),
    'NW': _('Nidwalden'),
    'OW': _('Obwalden'),
    'SH': _('Schaffhausen'),
    'SZ': _('Schwyz'),
    'SO': _('Solothurn'),
    'SG': _('St. Gallen'),
    'TG': _('Thurgau'),
    'TI': _('Ticino'),
    'UR': _('Uri'),
    'VS': _('Valais'),
    'VD': _('Vaud'),
    'ZG': _('Zug'),
    'ZH': _('Zürich'),
}


class SwissHolidays(object):
    """ Provides the ability to check dates against Swiss holidays and to
    list the holidays for a given year. Builds on the code from
    python-holidays.

    See the ``Switzerland`` class here:
    https://github.com/dr-prodigy/python-holidays/blob/master/holidays.py

    Supports the following features which are not provided by python-holidays:

    * Translatable strings.
    * The ability to combine cantons.
    * The ability to add extra holidays for specific days of any year.

    The interface is inspired by pyhton-holidays, thought he API surface
    is a bit smaller and there is less magic.

    """

    def __init__(self, cantons=(), other=(), timezone='Europe/Zurich'):
        self._cantons = set()
        self._other = defaultdict(set)

        for canton in cantons:
            self.add_canton(canton)

        for month, day, description in other:
            self.add_holiday(month, day, description)

    def add_canton(self, canton):
        assert canton in CANTONS

        self._cantons.add(canton)

    def add_holiday(self, month, day, description):
        assert 1 <= month <= 12
        assert 1 <= day <= 31

        self._other[(month, day)].add(description)

    def __bool__(self):
        return (self._cantons or self._other) and True or False

    def __contains__(self, dt):
        if not isinstance(dt, date) or isinstance(dt, datetime):
            raise ValueError(f"Unsupported type: {type(dt)}")

        if (dt.month, dt.day) in self._other:
            return True

        for holiday, descriptions in self.official(dt.year):
            if holiday == dt:
                return True

        return False

    def all(self, year):
        """ Returns all the holidays for the given year for the current
        set of cantons. If no cantons are selected and no other holidays
        are defined, nothing is returned.

        The result is a list in chronological order with each list entry
        being at tuple of date and at least one description (if multiple
        holidays are on a single day, a list of descriptions is returned).

        The list of descriptions is sorted alphabetically.

        """

        combined = defaultdict(set)

        for dt, descriptions in self.official(year):
            combined[dt].update(descriptions)

        for month, day in self._other:
            combined[date(year, month, day)] |= self._other[(month, day)]

        dates = sorted(list(combined.keys()))

        return [(dt, sorted(combined[dt])) for dt in dates]

    def between(self, start, end):
        """ Returns all the holidays between the given start and end date in
        the same manner ass :meth:`all`.

        """
        assert start <= end

        if start.year == end.year:
            years = (start.year, )
        else:
            years = (start.year, end.year)

        def generate():
            for year in years:
                for dt, descriptions in self.all(year):
                    if start <= dt and dt <= end:
                        yield dt, descriptions

        return list(generate())

    def other(self, year):
        """ Returns all custom defined holidays for the given year. """

        for month, day in self._other:
            yield date(year, month, day), self._other[(month, day)]

    def official(self, year):
        """ Like :meth:`all`, but only includes the official holidays,
        not the custom defined ones.

        If no cantons are selected, no official holidays are returned, not
        even the national ones.

        The description only ever contains a tuple with one item. This is
        for congruence with the other methods of this class, where the
        description is always in an iterable (not necessarily a tuple).

        """

        if not self._cantons:
            return

        yield date(year, 1, 1), (_("Neujahrestag"), )

        if self._cantons & {'AG', 'BE', 'FR', 'GE', 'GL', 'GR', 'JU', 'LU',
                            'NE', 'OW', 'SH', 'SO', 'TG', 'VD', 'ZG', 'ZH'}:

            yield date(year, 1, 2), (_("Berchtoldstag"), )

        if self._cantons & {'SZ', 'TI', 'UR'}:
            yield date(year, 1, 6), (_("Heilige Drei Könige"), )

        if self._cantons & {'NE'}:
            yield date(year, 3, 1), (
                _("Jahrestag der Ausrufung der Republik"), )

        if self._cantons & {'NW', 'SZ', 'TI', 'UR', 'VS'}:
            yield date(year, 3, 19), (_("Josefstag"), )

        if self._cantons & {'GL'} and year >= 1835:

            # First Thursday in April but not in Holy Week
            if date(year, 4, 1) + rd(weekday=FR) != easter(year) - rd(days=2):
                yield date(year, 4, 1) + rd(weekday=TH), (
                    _("Näfelser Fahrt"), )
            else:
                yield date(year, 4, 8) + rd(weekday=TH), (
                    _("Näfelser Fahrt"), )

        yield easter(year), (_("Ostern"), )

        # Good Friday is celebrated if we have a canton other than TI, VS
        if self._cantons > {'TI', 'VS'}:
            yield easter(year) - rd(days=2), (_("Karfreitag"), )

        # Easter Monday is celebrated if we have a canton other than VS
        if self._cantons > {'VS'}:
            yield easter(year) + rd(weekday=MO), (_("Ostermontag"), )

        if self._cantons & {
                'BL', 'BS', 'JU', 'NE', 'SH', 'SO', 'TG', 'TI', 'ZH'}:
            yield date(year, 5, 1), (_("Tag der Arbeit"), )

        yield easter(year) + rd(days=39), (_("Auffahrt"), )
        yield easter(year) + rd(days=49), (_("Pfingsten"), )
        yield easter(year) + rd(days=50), (_("Pfingstmontag"), )

        if self._cantons & {
                'AI', 'JU', 'LU', 'NW', 'OW', 'SZ', 'TI', 'UR', 'VS', 'ZG'}:
            yield easter(year) + rd(days=60), (_("Fronleichnam"), )

        if self._cantons & {'JU'}:
            yield date(year, 6, 23), (_("Fest der Unabhängigkeit"), )

        if self._cantons & {'TI'}:
            yield date(year, 6, 29), (_("Peter und Paul"), )

        if year >= 1994:
            yield date(year, 8, 1), (_("Nationalfeiertag"), )

        if self._cantons & {
                'AI', 'JU', 'LU', 'NW', 'OW', 'SZ', 'TI', 'UR', 'VS', 'ZG'}:
            yield date(year, 8, 15), (_("Mariä Himmelfahrt"), )

        if self._cantons & {'OW'}:
            yield date(year, 9, 25), (_("Bruder Klaus"), )

        if self._cantons & {
                'AI', 'GL', 'JU', 'LU', 'NW', 'OW', 'SG', 'SZ', 'TI', 'UR',
                'VS', 'ZG'}:
            yield date(year, 11, 1), (_("Allerheiligen"), )

        if self._cantons & {
                'AI', 'LU', 'NW', 'OW', 'SZ', 'TI', 'UR', 'VS', 'ZG'}:
            yield date(year, 12, 8), (_("Mariä Empfängnis"), )

        if self._cantons & {'GE'}:
            yield date(year, 12, 12), (_("Escalade de Genève"), )

        yield date(year, 12, 25), (_("Weihnachten"), )

        if self._cantons & {
                'AG', 'AR', 'AI', 'BL', 'BS', 'BE', 'FR', 'GL', 'GR', 'LU',
                'NE', 'NW', 'OW', 'SG', 'SH', 'SZ', 'SO', 'TG', 'TI', 'UR',
                'ZG', 'ZH'}:
            yield date(year, 12, 26), (_("Stephanstag"), )

        if self._cantons & {'GE'}:
            yield date(year, 12, 31), (_("Wiederherstellung der Republik"), )
